from __future__ import annotations

import logging
from dataclasses import dataclass, field
from functools import partial
from typing import Any, Callable, Dict, Iterable

from ..data import ParseCompound
from .shared import (
    MALFORMED_LINE_STARTS,
    _add_database_link,
    _add_type,
    _set_gibbs0,
    _set_name,
)

__all__ = ["CompoundParser"]

logger = logging.getLogger(__name__)


def _set_atom_charges(compounds: Dict[str, ParseCompound], id_: str, content: str) -> None:
    try:
        compounds[id_].charge += int(content[1:-1].split()[-1])
    except ValueError:  # conversion failed
        pass


def _set_chemical_formula(compounds: Dict[str, ParseCompound], id_: str, content: str) -> None:
    atom, count = content[1:-1].split(" ")
    try:
        compounds[id_].formula[atom] = int(count)
    except ValueError:  # conversion failed
        pass


def _set_smiles(compounds: Dict[str, ParseCompound], id_: str, content: str) -> None:
    compounds[id_].smiles = content


@dataclass
class CompoundParser:
    type_map: dict[str, str]
    actions: Dict[str, Callable[[Any, Any, Any], None]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.actions = {
            "TYPES": partial(_add_type, type_map=self.type_map),
            "COMMON-NAME": _set_name,
            "ATOM-CHARGES": _set_atom_charges,
            "CHEMICAL-FORMULA": _set_chemical_formula,
            "DBLINKS": _add_database_link,
            "GIBBS-0": _set_gibbs0,
            "SMILES": _set_smiles,
        }

    def parse(self, file: Iterable[str]) -> Dict[str, ParseCompound]:
        """Parse."""
        compounds: Dict[str, ParseCompound] = {}
        id_ = ""
        for line in file:
            if any(line.startswith(i) for i in MALFORMED_LINE_STARTS):
                continue
            try:
                identifier, content = line.rstrip().split(" - ", maxsplit=1)
            except ValueError:
                logger.info(f"Malformed line in compoudns.dat: {line}")
                continue

            if identifier == "UNIQUE-ID":
                base_id = content
                id_ = content + "_c"
                compounds[id_] = ParseCompound(
                    id=id_,
                    base_id=base_id,
                    compartment="CYTOSOL",
                )
            else:
                if (action := self.actions.get(identifier, None)) is not None:
                    action(compounds, id_, content)
        return compounds



from typing import Dict

from ..data import ParseReaction


def fix_set_reaction_stoichiometry(parse_reactions: Dict[str, ParseReaction]) -> Dict[str, ParseReaction]:
    """Set the stoichiometry from the information given by the substrates and products."""
    new_reactions = {}
    for rxn_id, reaction in parse_reactions.items():
        substrates = reaction.substrates
        products = reaction.products

        # Check for duplicates
        for compound in set(substrates).intersection(set(products)):
            diff = products[compound] - abs(substrates[compound])
            if diff == 0:
                del substrates[compound]
                del products[compound]
            elif diff < 0:
                substrates[compound] = diff
                del products[compound]
            else:
                del substrates[compound]
                products[compound] = diff

        # Create stoichiometry
        stoichiometries = {**substrates, **products}
        if len(stoichiometries) > 1:
            reaction.stoichiometries = stoichiometries
            new_reactions[rxn_id] = reaction
    return new_reactions



from __future__ import annotations

from typing import Dict

from ....core.reaction import KineticData
from ..data import ParseReaction


def fix_kinetic_data(rxns: Dict[str, ParseReaction]) -> Dict[str, ParseReaction]:
    for rxn in rxns.values():
        rxn.kinetic_data = {
            k: KineticData(km=v.get("km", {}), kcat=v.get("kcat", {})) for k, v in rxn.enzrxns.items()
        }
        # del rxn.enzrxns
    return rxns



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Set, Tuple

from ...core.reaction import KineticData, Monomer

__all__ = [
    "ParseGene",
    "ParseCompound",
    "ParseReaction",
    "ParseEnzyme",
]


@dataclass
class ParseGene:
    id: str
    product: str | None = None
    database_links: Dict[str, Set[str]] = field(default_factory=dict)


@dataclass
class ParseCompound:
    id: str
    base_id: str
    charge: int = 0
    compartment: str = "CYTOSOL"
    smiles: str | None = None
    name: str | None = None
    gibbs0: float | None = None
    types: List[str] = field(default_factory=list)
    formula: Dict[str, float] = field(default_factory=dict)
    database_links: Dict[str, Set[str]] = field(default_factory=dict)


@dataclass
class ParseReaction:
    id: str
    base_id: str
    bounds: Tuple[float, float] = (0, 0)
    name: str | None = None
    ec: str | None = None
    gibbs0: float | None = None
    direction: str = "LEFT-TO-RIGHT"
    reversible: bool = False
    transmembrane: bool = False
    substrates: Dict[str, float] = field(default_factory=dict)
    substrate_compartments: Dict[str, str] = field(default_factory=dict)
    products: Dict[str, float] = field(default_factory=dict)
    product_compartments: Dict[str, str] = field(default_factory=dict)
    types: List[str] = field(default_factory=list)
    locations: List[str] = field(default_factory=list)
    pathways: Set[str] = field(default_factory=set)
    enzymes: Set[str] = field(default_factory=set)
    database_links: Dict[str, Set[str]] = field(default_factory=dict)
    _var: int | None = None
    # Later additions
    monomers: Dict[str, Set[str]] = field(default_factory=dict)
    monomers_annotated: Dict[str, Dict[str, Monomer]] = field(default_factory=dict)
    sequences: Dict[str, str] = field(default_factory=dict)
    enzrxns: Dict[str, Dict[str, Dict[str, float]]] = field(default_factory=dict)
    kinetic_data: Dict[str, KineticData] = field(default_factory=dict)
    compartment: str | Tuple[str, str] | None = None
    stoichiometries: Dict[str, float] = field(default_factory=dict)


@dataclass
class ParseEnzyme:
    id: str
    enzyme: str | None = None
    kcat: Dict[str, float] = field(default_factory=dict)
    km: Dict[str, float] = field(default_factory=dict)
    vmax: Dict[str, float] = field(default_factory=dict)



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterable

from ..data import ParseEnzyme
from .shared import MALFORMED_LINE_STARTS

__all__ = ["EnzymeParser"]

logger = logging.getLogger(__name__)


def _set_enzyme(enzrxns: Dict[str, ParseEnzyme], id_: str, enzyme: str) -> None:
    enzrxns[id_].enzyme = enzyme


def _add_kcat(enzrxns: Dict[str, ParseEnzyme], id_: str, substrate: str, kcat: str) -> None:
    try:
        enzrxns[id_].kcat.setdefault(substrate, float(kcat))
    except ValueError:  # conversion failed
        pass


def _add_km(enzrxns: Dict[str, ParseEnzyme], id_: str, substrate: str, km: str) -> None:
    try:
        enzrxns[id_].km.setdefault(substrate, float(km))
    except ValueError:  # conversion failed
        pass


def _add_vmax(enzrxns: Dict[str, ParseEnzyme], id_: str, substrate: str, vmax: str) -> None:
    try:
        enzrxns[id_].vmax.setdefault(substrate, float(vmax))
    except ValueError:  # conversion failed
        pass


@dataclass
class EnzymeParser:
    actions: Dict[str, Callable[[Any, Any, Any], None]] = field(default_factory=dict)
    sub_actions: Dict[str, Dict[str, Callable[[Any, Any, Any, Any], None]]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.actions = {
            "ENZYME": _set_enzyme,
        }
        self.sub_actions = {
            "^SUBSTRATE": {"KM": _add_km, "VMAX": _add_vmax, "KCAT": _add_kcat},
        }

    def parse(self, file: Iterable[str]) -> Dict[str, ParseEnzyme]:
        """Parse."""
        id_ = ""
        enzrxns = {}
        last_identifier = ""
        last_content = ""
        for line in file:
            if any(line.startswith(i) for i in MALFORMED_LINE_STARTS):
                continue
            try:
                identifier, content = line.rstrip().split(" - ", maxsplit=1)
            except ValueError:
                logger.info(f"Malformed line in enzymes.dat {line}")
                continue

            if identifier == "UNIQUE-ID":
                id_ = content
                enzrxns[id_] = ParseEnzyme(id=id_)
            elif identifier.startswith("^"):
                if (subaction := self.sub_actions.get(identifier, {}).get(last_identifier, None)) is not None:
                    subaction(enzrxns, id_, content, last_content)
            else:
                if (action := self.actions.get(identifier, None)) is not None:
                    action(enzrxns, id_, content)
                    last_identifier = identifier
                    last_content = content
        return enzrxns



from __future__ import annotations

import itertools as it
import copy
from typing import Dict, List

from ..data import ParseCompound, ParseReaction
from ..utils import _reaction_is_bad


def _get_compound_variants(
    cpds: Dict[str, float],
    cpd_types: Dict[str, List[str]],
) -> Dict[str, List[str]]:
    return {cpd: cpd_types[cpd] for cpd in cpds if cpd in cpd_types}


def _get_reaction_variant(
    rxn: ParseReaction,
    compound_map: dict[str, str],
    count: int,
) -> ParseReaction:
    local = copy.copy(rxn)
    local.substrates = {compound_map.get(k, k): v for k, v in local.substrates.items()}
    local.products = {compound_map.get(k, k): v for k, v in local.products.items()}
    local.substrate_compartments = {
        compound_map.get(k, k): v for k, v in local.substrate_compartments.items()
    }
    local.product_compartments = {compound_map.get(k, k): v for k, v in local.product_compartments.items()}
    local.id = f"{local.id}__var__{count}"
    local._var = count
    return local


def fix_create_reaction_variants(
    rxns: Dict[str, ParseReaction],
    cpds: Dict[str, ParseCompound],
    compound_types: Dict[str, List[str]],
) -> Dict[str, ParseReaction]:
    """Create all mass and charge balanced reaction variants of reactions containing compound classes."""
    new_rxns = {}
    for rxn_id, rxn in rxns.items():
        count = 0
        substrate_variants = _get_compound_variants(rxn.substrates, compound_types)
        product_variants = _get_compound_variants(rxn.products, compound_types)

        variants = {**substrate_variants, **product_variants}
        if len(variants) == 0:
            new_rxns[rxn_id] = rxn
        else:
            for new_cpds, old_cpds in zip(
                it.product(*variants.values()),
                it.repeat(variants.keys()),
            ):
                compound_map = dict(zip(old_cpds, new_cpds))
                new_rxn = _get_reaction_variant(rxn, compound_map, count)
                # Performance improvement: filter garbage reactions already here
                if _reaction_is_bad(new_rxn, cpds):
                    continue
                new_rxns[new_rxn.id] = new_rxn
                count += 1
    return new_rxns



from __future__ import annotations

from typing import Dict

from ....core.reaction import Monomer
from ..data import ParseReaction


def fix_annotate_monomers(
    rxns: Dict[str, ParseReaction], monomers: Dict[str, Monomer]
) -> Dict[str, ParseReaction]:
    for rxn in rxns.values():
        rxn.monomers_annotated = {
            k: {monomer: monomers[monomer] for monomer in v if monomer in monomers}
            for k, v in rxn.monomers.items()
        }
        del rxn.monomers
    return rxns



from __future__ import annotations

import re
from typing import Dict, Iterator

__all__ = ["SequenceParser"]


class SequenceParser:
    def parse(self, file: Iterator[str]) -> Dict[str, str]:
        RE_PAT = re.compile(r"^>gnl\|.*?\|")
        sequences: Dict[str, str] = {}
        while True:
            try:
                id_ = re.sub(RE_PAT, "", next(file)).split(" ", maxsplit=1)[0]
                sequence = next(file).strip()
                sequences[id_] = sequence
            except StopIteration:
                break
        return sequences
